/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Parallel Task Manager (PTM) for Heterogenous Computing System */
/*                     Version 1.0 (DEC-1994)                    */
/*                          HUI Chun-Fai                         */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* * * * * * * * * */
/* Filename: ptm.c */
/* * * * * * * * * */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* The program manages parallel tasks running on a heterogenous  */
/* computing system formed with PVM (Parallel Virtual Machine)   */
/* Version 3.3.                                                  */
/*                                                               */
/* The program should be compiled with the header file 'ptm.h'   */
/* the same path with the following commands:                    */
/*                                                               */
/* SUN Workstation:                                              */
/*    cc -g -I$R/include -o ptm ptm.c $R/lib/$A/libpvm3.a        */
/*                                                               */
/* MASPAR:                                                       */
/*    mpl_cc -g -I$R/include -o ptm ptm.c $R/lib/$A/libpvm3.a    */
/*                                                               */
/* PARAGON:                                                      */
/*    icc -g -I$R/include -o ptm ptm.c $R/lib/$A/libpvm3.a -lrpc */
/*                                                               */
/*    where $R and $A stand for $PVM_ROOT and $PVM_ARCH          */
/*                                                               */
/* Before running ptm, PVM should be start first by typing:      */
/*    % pvmd hostfile                                            */
/*                                                               */
/* After that, type:                                             */
/*    % ptm tgd_file                                             */
/*                                                               */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include "pvm3.h" /* the pvm3 header file */
#include "ptm.h" /* the header file for ptm */

main(argc,argv) /* first argument = tgd filename */
int argc;
char **argv;
{
   FILE *TGDFile; /* pointer to tgd pointer */
   FILE *LogFile; /* pointer to log file */
   int i,j,k,l,Cntr; /* general counters */
   int ReadMode; /* indicator: task or variable description line */
   int ArgCntr, TaskCntr=0, VarCntr=0; /* counters: arg, task, var */
   int OutArg=0; /* indicator: output or input parameter */
   char String[MAXSTR]; /* buffer for input string from tgd file */
   int Scheduled[MAXTASK]; /* task scheduled */
   int Phase,Task, match; /* phase, task pointers */
   int NScheduled; /* number of tasks being scheduled */
   int VarFreed[MAXVAR]; /* variables to be freed */
   int hTask,hVar,hSize; /* task, var pointers */
   int Where, hNumt; /* where to spawn the task, number of tasks spawned */
   int TypeError; /* variable type error indicator */
   struct timeval hTV; /* for measuring elapsed time of task */
   struct timezone hTZ; /* for measuring elapsed time of task */

   printf("\nPARALLEL TASK MANAGER for Heterogeneous Computing System\n");
   printf("PTM V1.0 (DEC-1994)\n");
   printf("Developer: HUI Chun-FAI\n\n");

   /* if no tgd filename is specified, 'default.tgd' will be read */
   if (argc<2)
      TGDFile = fopen("default.tgd","r");
   else
      TGDFile = fopen(argv[1],"r");

   /* No tgd filename is specified and 'default.tgd' cannot be found */
   if (TGDFile == NULL) { 
      printf("[PTM]: Input File Error\n");
      printf("Filename not specified and 'default.tgd' not found!\n");
      printf("Usage: % ptm tgd_file log_file\n");
      exit(1);
      }

   printf("[PTM]: Stage 1, reading task graph description . . .\n");

/* * * * * * * * * * * * * * * */
/* READING OF TGD FILE STARTS  */
/* * * * * * * * * * * * * * * */

   while (!feof(TGDFile)) { /* start reading TGD file */

      fscanf(TGDFile,"%s",String);

      if (!strcmp(String,"[TASK]")) { /* check if task section begins */
	 ReadMode = 1;
	 ArgCntr = 0;
	 continue;
      }

      if (!strcmp(String,"[VAR]")) { /* check if variable section begins */
	 ReadMode = 0;
	 ArgCntr = 0;
	 continue;
      }

      if (!strcmp(String,"[END]")) { /* check if end of TGD file */
	 printf("[END]\n");
	 break;
      }

      if (!strcmp(String,";")) { /* check if end of description line */

	 if (ReadMode) { /* if task section is being read */
	   TaskCntr++;
	   OutArg=0;
	 }

	 else { /* if variable section is being read */
	    /* only 2 arguments are given in variable description line */
	    if (ArgCntr==2) { 
	       hVAR[VarCntr].rows = 1; /* assume number of row  = 1 */
	       hVAR[VarCntr].cols = 0; /* assume number of column = 0 */
	    }
	    /* only 3 arguments are given in variable description line */ 
	    if (ArgCntr==3) { 
	       hVAR[VarCntr].cols = 0; /* assume number of column = 0 */
	    }

	    VarCntr++;
	 }

	 ArgCntr = 0;
	 continue;
      }

      /* ReadMode = [TASK], task description lines are being read */ 
      if (ReadMode==1) { 

	 if (ArgCntr==0) { /* the first argument = task name */
	    strcpy(hTASK[TaskCntr].name,String);
	    hTASK[TaskCntr].ocntr = 0;
	    hTASK[TaskCntr].icntr = 0;
	    hTASK[TaskCntr].fcntr = 0;
	 }

	 if (ArgCntr==1) /* number of copies of task to be spawned */
	    hTASK[TaskCntr].nproc = atoi(String);

	 if (ArgCntr==2) { /* the host name */
	    strcpy(hTASK[TaskCntr].host,String);
	       /* first character of host name is capital */
	       if (isupper(String[0]))  
		  /* spawn task to arch */
		  hTASK[TaskCntr].arch = 1;  
	       else /* first character is small */
		  /* spawn task to particular host */      
		  hTASK[TaskCntr].arch = 0; 
	 }

	 if (ArgCntr>2) { /* the input arguments and output results */

	    if (!strcmp(String,"/")) { /* check if output results start */
	       OutArg = 1;
	    }

	    else {

	       if (OutArg) { /* to get output results of the task */
		  for (i=0;i<VarCntr;i++) { /* check variable list */
		     if (!strcmp(String,hVAR[i].name)) { /* found it */
			Cntr = hTASK[TaskCntr].ocntr;
			hTASK[TaskCntr].output[Cntr] = i;
			hTASK[TaskCntr].ocntr++;
			break;
		     }
		  }
		  if (i>=VarCntr) { /* if the variable not valid */
		     printf("[PTM]: Output Parameter Error\n");
		     printf("at %dth line of Task Section: '%s'\n", TaskCntr+1,hTASK[TaskCntr].name);
		     exit(1);
		  }
	       }

	       else { /* to get input arguments of the task */
		  for (i=0;i<VarCntr;i++) { /* check variable list */
		     if (!strcmp(String,hVAR[i].name)) { /* found it */
			Cntr = hTASK[TaskCntr].icntr;
			hTASK[TaskCntr].input[Cntr] = i;
			hTASK[TaskCntr].icntr++;
			if (hVAR[i].type==NAME) /* a filename */
			   hTASK[TaskCntr].fcntr++;
			break;
		     }
		  }
		  if (i>=VarCntr) {
		     printf("[PTM]: Input Parameter Error\n");
		     printf("at %dth line of Task Section: '%s'\n", TaskCntr+1,hTASK[TaskCntr].name);
		     exit(1);
		  }
	       }
	    }
	 }

	 ArgCntr++;

      }

      if (ReadMode==0) { /* ReadMode = [VAR] */

	 if (ArgCntr==0) /* the first argument = var name */
	    strcpy(hVAR[VarCntr].name,String);

	 if (ArgCntr==1) { /* the var type */
	    TypeError = 1;
	    if (!strcmp(String,"i")) {
	       hVAR[VarCntr].type = INT;
	       TypeError = 0;
	    }
	    if (!strcmp(String,"r")) {
	       hVAR[VarCntr].type = REAL;
	       TypeError = 0;
	    }
	    if (!strcmp(String,"c")) {
	       hVAR[VarCntr].type = CHAR;
	       TypeError = 0;
	    }
	    if (!strcmp(String,"f")) {
	       hVAR[VarCntr].type = NAME;
	       TypeError = 0;
	    }
	    if (TypeError) {
	       printf("[PTM]: Variable Type Error\n");
	       printf("at %dth line of Var Section: '%s'\n", VarCntr+1,hVAR[VarCntr].name);
	       exit(1);
	    }
	 }

	 if (ArgCntr==2) { /* first dimension of array or filename */

	    if (hVAR[VarCntr].type == NAME) {
	       strcpy(hVAR[VarCntr].cptr,String);
	       hVAR[VarCntr].rows = strlen(String);
	       hVAR[VarCntr].cols = 0;
	    }

	    if (hVAR[VarCntr].type == CHAR) {
	       hVAR[VarCntr].rows = atoi(String);
	       hVAR[VarCntr].cols = 0;
	    }

	    if (hVAR[VarCntr].type==INT || hVAR[VarCntr].type==REAL) {
	       hVAR[VarCntr].rows = atoi(String);
	    }
	 }

	 if (ArgCntr==3) { /* second dimension of array */

	    if (hVAR[VarCntr].type==INT || hVAR[VarCntr].type==REAL) {
	       hVAR[VarCntr].cols = atoi(String);
	    }

	 }

	 if (ArgCntr>3) {
	    printf("[PTM]: Variable Definition Error\n");
	    printf("at %dth line of Var Section: '%s'\n", VarCntr+1,hVAR[VarCntr].name);
	    exit(1);
	 }

	 ArgCntr++;

      }

   } /* while (!feof) */

   fclose(TGDFile);

/* * * * * * * * * * * * * * * * */
/* READING OF TGD FILE COMPLETED */
/* * * * * * * * * * * * * * * * */

   printf("Variables:\n");

   for (i=0;i<VarCntr;i++) {
      printf("VAR #%d: ",i);
      printf("%s,",hVAR[i].name);
      printf("%d,",hVAR[i].type);
      printf("%d,",hVAR[i].rows);
      printf("%d,",hVAR[i].cols);
      if (hVAR[i].type == NAME) printf("%s\n",hVAR[i].cptr);
      else printf("\n");
   }

   printf("Tasks:\n");

   for (i=0;i<TaskCntr;i++) {
      printf("TASK #%d: ",i);
      printf("%s,",hTASK[i].name);
      printf("%s with inputs: ",hTASK[i].host);
      for (Cntr=0;Cntr<hTASK[i].icntr;Cntr++)
	 printf("%d,",hTASK[i].input[Cntr]);
      printf(" outputs: ");
      for (Cntr=0;Cntr<hTASK[i].ocntr;Cntr++)
	 printf("%d,",hTASK[i].output[Cntr]);
      printf(" and ");
      printf("%d filenames\n",hTASK[i].fcntr);
   }

/* * * * * * * * * * * * * * * */
/* SCHEDULING OF TASKS STARTS  */
/* * * * * * * * * * * * * * * */

/* initial phase */

   printf("[PTM]: Stage 2, scheduling of tasks . . .\n");

   for (Task=0;Task<TaskCntr;Task++)
      Scheduled[Task] = 0;

   NScheduled = 0; /* set number of tasks scheduled to 0 */
   hPHASE[0].tcntr = 0; /* set number of tasks scheduled to 0 */
   hPHASE[0].rcntr = 0; /* set number of results to 0 */

   /* Initial phase of scheduling: to select all substantive tasks */
   for (Task=0;Task<TaskCntr;Task++) { 
      if ((hTASK[Task].icntr-hTASK[Task].fcntr)==0) { /* substantive task */
	 hPHASE[0].task[hPHASE[0].tcntr] = Task;
	 Scheduled[Task] = 1;
	 NScheduled++;
	 hPHASE[0].tcntr++;
	 printf("'%s' scheduled in phase 0.\n",hTASK[Task].name);
	 for (i=0;i<hTASK[Task].ocntr;i++) {
	    hPHASE[0].result[hPHASE[0].rcntr] = hTASK[Task].output[i];
	    hPHASE[0].rcntr++;
	 }
      }
   }

   if (hPHASE[0].tcntr==0) {
      printf("[PTM]: Task Scheduling Error\n");
      printf("No substantive task can be found in initial phase!\n");
      exit(1);
   }

   /* Subsequent phases of shceduling */
   /* Task is selected if all data it requires is ready in the current phase */
   Phase = 1;
   while (NScheduled<TaskCntr) { /* schedule tasks until all scheduled */
      hPHASE[Phase].tcntr = 0;
      hPHASE[Phase].rcntr = 0;
      for (Task=0;Task<TaskCntr;Task++) { /* search all tasks */
	 if (!Scheduled[Task]) { /* find a task not scheduled */
	    /* find if all data it requires are ready */
	    for (i=0;i<hTASK[Task].icntr;i++) {
	       if (hVAR[hTASK[Task].input[i]].type==NAME)
		  continue;
	       match = 0;
	       for (j=0;j<hPHASE[Phase-1].rcntr;j++) {
		  if (hPHASE[Phase-1].result[j]==hTASK[Task].input[i]) {
		     match = 1;
		     break;
		  }
	       }
	       if (!match) {
		  break;
	       }
	    }
	    if (match) { /* all input parameters are available */
	       hPHASE[Phase].task[hPHASE[Phase].tcntr] = Task;
	       Scheduled[Task] = 1;
	       NScheduled++;
	       hPHASE[Phase].tcntr++;
	       printf("'%s' scheduled in phase %d.\n",hTASK[Task].name,Phase);
	       for (i=0;i<hTASK[Task].ocntr;i++) {
		  hPHASE[Phase].result[hPHASE[Phase].rcntr] = hTASK[Task].output[i];
		  hPHASE[Phase].rcntr++;
	       }
	    }
	 }
      }
      if (hPHASE[Phase].tcntr==0) {
	 printf("[PTM]: Task Scheduling Error\n");
	 printf("No task can be scheduled in Phase %d!\n", Phase);
	 exit(1);
      }
      else {
	 for (i=0;i<hPHASE[Phase-1].rcntr;i++)
	    hPHASE[Phase].result[hPHASE[Phase].rcntr++] = hPHASE[Phase-1].result[i];
	 Phase++;
      }
   }

   for (i=0;i<VarCntr;i++) {
      VarFreed[i] = 0; /* variables to be freed */
   }

   /* Search backwards from last phase to find which variables can be freed */
   for (i=Phase-1;i>=0;i--) {
      hPHASE[i].dcntr = 0;
      for (Task=0;Task<hPHASE[i].tcntr;Task++) {
	 for (j=0;j<hTASK[hPHASE[i].task[Task]].icntr;j++) {
	    if (!VarFreed[hTASK[hPHASE[i].task[Task]].input[j]]) {
	       VarFreed[hTASK[hPHASE[i].task[Task]].input[j]] = 1;
	       hPHASE[i].dealloc[hPHASE[i].dcntr] = hTASK[hPHASE[i].task[Task]].input[j];
	       hPHASE[i].dcntr++;
	    }
	 }
      }
   }

/* * * * * * * * * * * * * * * * */
/* SCHEDULING OF TASKS COMPLETED */
/* * * * * * * * * * * * * * * * */


/* * * * * * * * * * * * * * * */
/* DESPATCHING OF TASKS STARTS */
/* * * * * * * * * * * * * * * */

   LogFile = fopen("ptm.log","w"); /* open log file */

   hTZ.tz_minuteswest = 0;
   hTZ.tz_dsttime = 0;
   gettimeofday(&hTV,&hTZ); /* get starting time of application */
   hSTARTTIME = (double)hTV.tv_sec + (double)hTV.tv_usec/1000000.0;
   
   for (i=0;i<Phase;i++) { /* Start tasks in each phase */
      fprintf(LogFile,"*** PHASE %d ***\n\n",i);
      printf("[PTM]: Phase %d\n",i);
      for (j=0;j<hPHASE[i].tcntr;j++) {
	 hTask = hPHASE[i].task[j];
	 printf (" * Task %d\n",hTask);
	 hARGC = hTASK[hTask].icntr;
	 hNPROC = hTASK[hTask].nproc;
	 printf ("   - %d arguments\n",hARGC);
	 for (k=0;k<hTASK[hTask].icntr;k++) {
	    hVar = hTASK[hTask].input[k];
	    printf("   - Var %d, %s: ",hVar,hVAR[hVar].name);
	    hARGV[k].type = hVAR[hVar].type;
	    hARGV[k].rows = hVAR[hVar].rows;
	    hARGV[k].cols = hVAR[hVar].cols;
	    printf ("type=%d,rows=%d,cols=%d\n",hARGV[k].type,hARGV[k].rows,hARGV[k].cols);
	    switch (hVAR[hVar].type) {
	    case INT:
	       hARGV[k].iptr = hVAR[hVar].iptr;
	       break;
	    case REAL:
	       hARGV[k].rptr = hVAR[hVar].rptr;
	       break;
	    case CHAR:
	       strcpy(hARGV[k].cptr,hVAR[hVar].cptr);
	       break;
	    case NAME:
	       strcpy(hARGV[k].cptr,hVAR[hVar].cptr);
	    }
	 }

	 if (hTASK[hTask].arch==1)
	    Where = PvmTaskArch;
	 else
	    Where = PvmTaskHost;
	 hNumt=pvm_spawn(hTASK[hTask].name,(char**)0,Where,hTASK[hTask].host,hNPROC,hTIDS);
	 printf("   - %d copies spawned\n",hNumt);
	 if (hNumt<hNPROC) {
	    printf("[PTM]: Task Starting Error\n");
	    switch(hNumt) {
	    case PvmBadParam:
	       printf("Giving an invalid argument value\n");
	       break;
	    case PvmNoHost:
	       printf("Specified host is not in the virtual machine\n");
	       break;
	    case PvmNoFile:
	       printf("Specified executable cannot be found\n");
	       printf("File should be in ~/pvm3/bin/ARCH\n");
	       break;
	    case PvmNoMem:
	       printf("No enough memory on host\n");
	       break;
	    case PvmSysErr:
	       printf("pvmd not responding\n");
	       break;
	    case PvmOutOfRes:
	       printf("Out of resources\n");
	    }
	    exit(1);
	 }
	 hTASK[hTask].tid = hTIDS[0];
	 hTASK[hTask].status = RUNNING;
         gettimeofday(&hTV,&hTZ); /* starting time of task */
         hTASK[hTask].start = (double)hTV.tv_sec+(double)hTV.tv_usec/1000000.0;
	 hSEND();

	 printf("   - arguments sent \n");
      }

      printf("[PTM]: All tasks for this phase spawned\n\n");

      for (j=0;j<hPHASE[i].tcntr;j++) {
	 hRECV();

         gettimeofday(&hTV,&hTZ); /* ending time of task */

	 for (k=0;k<hPHASE[i].tcntr;k++) {
	    hTask = hPHASE[i].task[k];
	    if (hTID==hTASK[hTask].tid)
	       break;
	 }
	 if (k>=hPHASE[i].tcntr) {
	    printf("[PTM]: Unknown Task Error\n");
	    printf("Message received from an unknown task in phase %d!\n", Phase);
	    exit(1);
	 }
         printf(" * Task %d\n",hTask);
         printf("   - results returned\n");

         hTASK[hTask].end = (double)hTV.tv_sec+(double)hTV.tv_usec/1000000.0; 
         hTIME = hTASK[hTask].end-hTASK[hTask].start;

         fprintf(LogFile,"Task %d completed. Time taken = %f s\n",hTask,hTIME);

	 for (k=0;k<hTASK[hTask].ocntr;k++) { /* receive results from task */
	    hVar = hTASK[hTask].output[k];
	    printf("   - Var %d, %s: ",hVar,hVAR[hVar].name);
	    printf ("type=%d,rows=%d,cols=%d\n",hVAR[hVar].type,hVAR[hVar].rows,hVAR[hVar].cols);
	    switch (hRETV[k].type) {
	    case INT:
		hVAR[hVar].iptr = hRETV[k].iptr;
		break;
	    case REAL:
		hVAR[hVar].rptr = hRETV[k].rptr;
		break;
	    case CHAR:
	       strcpy(hVAR[hVar].cptr,hRETV[k].cptr);
	    }
	 }
      printf("\n");
      }

      for (j=0;j<hPHASE[i].dcntr;j++) { /* free unnecessary variables */
	 hVar = hPHASE[i].dealloc[j];
	 fprintf(LogFile,"VAR:%s\n",hVAR[hVar].name);
	 if (hVAR[hVar].cols) { /* if cols != 0 the variable is 2-dimensional */
	    hSize = hVAR[hVar].rows * hVAR[hVar].cols;
	    switch (hVAR[hVar].type) {
	    case INT:
	       for (k=0;k<hVAR[hVar].rows;k++) {
		  for (l=0;l<hVAR[hVar].cols;l++) {
		     fprintf(LogFile,"%d ",hVAR[hVar].iptr[k*hVAR[hVar].cols+l]);
		  }
		  fprintf(LogFile,"\n");
	       }
	       break;
	    case REAL:
	       for (k=0;k<hVAR[hVar].rows;k++) {
		  for (l=0;l<hVAR[hVar].cols;l++) {
		     fprintf(LogFile,"%d ",hVAR[hVar].rptr[k*hVAR[hVar].cols+l]);
		  }
		  fprintf(LogFile,"\n");
	       }
	    }
	 }
	 else { /* the variable is 1-dimensional */
	    hSize = hVAR[hVar].rows;
	    switch (hVAR[hVar].type) {
	    case INT:
	       for (k=0;k<hSize;k++)
		  fprintf(LogFile,"%d ",hVAR[hVar].iptr[k]);
	       break;
	    case REAL:
	       for (k=0;k<hSize;k++)
		  fprintf(LogFile,"%f ",hVAR[hVar].rptr[k]);
	       break;
	    case CHAR:
	       fprintf(LogFile,"%s",hVAR[hVar].cptr);
	       break;
	    case NAME:
	       fprintf(LogFile,"%s",hVAR[hVar].cptr);
	    }
	 }
	 fprintf(LogFile,"\n\n");
	 switch (hVAR[hVar].type) {
	 case INT:
	    free(hVAR[hVar].iptr);
	    break;
	 case REAL:
	    free(hVAR[hVar].rptr);
	 }
      hVAR[hVar].type = -1; /* freed variable */
      }
   }

   printf("[PTM]: ALL TASKS COMPLETED\n");

   fprintf(LogFile,"*** ALL TASKS COMPLETED *** \n\n");

   gettimeofday(&hTV,&hTZ);
   hENDTIME = (double)hTV.tv_sec+(double)hTV.tv_usec/1000000.0;
   hTIME = hENDTIME-hSTARTTIME;
   fprintf(LogFile,"Time taken = %f s\n",hTIME);

/* * * * * * * * * * * * * * * * * */
/* DESPATCHING OF TASKS COMPLETED  */
/* * * * * * * * * * * * * * * * * */


/* * * * * * * * * * * * * * * * * * * * * * * */
/* PRINTING OF REMAINING VARIABLES TO LOG FILE */
/* * * * * * * * * * * * * * * * * * * * * * * */
   
   for (i=0;i<VarCntr;i++) { /* print remaining variables to log file */
      hVar = i;
      if (hVAR[hVar].type >= 0) {
	 fprintf(LogFile,"** VAR:%s\n",hVAR[hVar].name);
	 if (hVAR[hVar].cols) {
	    hSize = hVAR[hVar].rows * hVAR[hVar].cols;
	    switch (hVAR[hVar].type) {
	    case INT:
	       for (k=0;k<hVAR[hVar].rows;k++) {
		  for (l=0;l<hVAR[hVar].cols;l++)
		     fprintf(LogFile,"%d ",hVAR[hVar].iptr[k*hVAR[hVar].cols+l]);
		  fprintf(LogFile,"\n");
	       }
	       break;
	    case REAL:
	       for (k=0;k<hVAR[hVar].rows;k++) {
		  for (l=0;l<hVAR[hVar].cols;l++)
		     fprintf(LogFile,"%d ",hVAR[hVar].rptr[k*hVAR[hVar].cols+l]);
		  fprintf(LogFile,"\n");
	       }
	    }
	 }
	 else {
	    hSize = hVAR[hVar].rows;
	    switch (hVAR[hVar].type) {
	    case INT:
	       for (k=0;k<hSize;k++)
		  fprintf(LogFile,"%d ",hVAR[hVar].iptr[k]);
	       break;
	    case REAL:
	       for (k=0;k<hSize;k++)
		  fprintf(LogFile,"%f ",hVAR[hVar].rptr[k]);
	       break;
	    case CHAR:
	       fprintf(LogFile,"%s",hVAR[hVar].cptr);
	       break;
	    case NAME:
	       fprintf(LogFile,"%s",hVAR[hVar].cptr);
	    }
	 }
	 fprintf(LogFile,"\n\n");
      }
   }
   fclose(LogFile);
   printf("\nN.B.:\n");
   printf("Please examine the log file 'ptm.log'\n");
   printf("   for values of variables in each phase.\n"); 
}


/* * * * * * * * * * * * */
/* END OF PROGRAM ptm.c  */
/* * * * * * * * * * * * */
